/*
 * Awake File: Easy file upload & download through HTTP with Java
 * Awake SQL: Remote JDBC access through HTTP.                                    
 * Copyright (C) 2011, Kawan Softwares S.A.S.
 * (http://www.awakeframework.org). All rights reserved.                                
 *                                                                               
 * Awake File/SQL is free software; you can redistribute it and/or                 
 * modify it under the terms of the GNU Lesser General Public                    
 * License as published by the Free Software Foundation; either                  
 * version 2.1 of the License, or (at your option) any later version.            
 *                                                                               
 * Awake File/SQL is distributed in the hope that it will be useful,               
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             
 * Lesser General Public License for more details.                               
 *                                                                               
 * You should have received a copy of the GNU Lesser General Public              
 * License along with this library; if not, write to the Free Software           
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301  USA
 *
 * Any modifications to this file must keep this entire header
 * intact.
 */

//TokenUtil.java
//Copyright (c) Kawan Softwares S.A.S, 2011
//
//Last Updates: 
// 18/07/11 19:05 NDP : TokenUtil: creation
// 30/09/11 15:25 NDP : TokenUtil: token is cut to 20 chars

package org.awakefw.file.api.util;

import org.apache.commons.lang3.StringUtils;
import org.awakefw.file.util.parms.Parameter;

/**
 * 
 * Recomputes a token value.
 * 
 * @author Nicolas de Pomereu
 */
public class TokenUtil {

    /**
     * Compute the token for the username and the computer values
     * 
     * @param username
     *            the username for login
     * @param secretForAuthToken
     *            the secret string to add to build the token
     * 
     * @return the computed access token
     * @throws Exception
     */
    public static String computeToken(String username, String secretForAuthToken)
	    throws Exception {
	// Add more secret info very hard to find for an external hacker

	StringBuilder tokenBuilder = new StringBuilder();
	tokenBuilder.append(System.getProperty("user.name"));
	tokenBuilder.append(ComputerInfo.getHostname());
	tokenBuilder.append(ComputerInfo.getMacAddress());

	if (secretForAuthToken != null) {
	    tokenBuilder.append(secretForAuthToken);
	}

	Sha1 hashcode = new Sha1();
	String token = hashcode
		.getHexHash((tokenBuilder.toString()).getBytes());

	token = StringUtils.left(token, Parameter.TOKEN_LEFT_SIZE);

	return token;
    }

}
