/*
 * Awake File: Easy file upload & download through HTTP with Java
 * Awake SQL: Remote JDBC access through HTTP.                                    
 * Copyright (C) 2011, Kawan Softwares S.A.S.
 * (http://www.awakeframework.org). All rights reserved.                                
 *                                                                               
 * Awake File/SQL is free software; you can redistribute it and/or                 
 * modify it under the terms of the GNU Lesser General Public                    
 * License as published by the Free Software Foundation; either                  
 * version 2.1 of the License, or (at your option) any later version.            
 *                                                                               
 * Awake File/SQL is distributed in the hope that it will be useful,               
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             
 * Lesser General Public License for more details.                               
 *                                                                               
 * You should have received a copy of the GNU Lesser General Public              
 * License along with this library; if not, write to the Free Software           
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301  USA
 *
 * Any modifications to this file must keep this entire header
 * intact.
 */

//ServerAwakeFileDispatch
//Copyright (c) Kawan Softwares S.A.S, 2011
//
//Last Updates: 
// 3 dc. 2009 19:48:16 Nicolas de Pomereu
// 22/09/09 16:20 NDP - Add Action.GET_FILE_LENGTH_ACTION
// 21/01/10 19:30 NDP - Add call to getCallableClassesWithoutAuthentication() for un-authenticated classes
// 25/01/10 19:30 NDP - If the file transfer operation is now allowed, return false when login
// 25/03/10 18:00 NDP - Remote call takes only one parameter for class+method name
// 06/04/10 15:50 NDP - ServerAwakeFileDispatch: Call getCallablePackages() 
//                      instead of getCallableMethodsWithoutAuthentication()
// 15/07/10 19:25 NDP - ServerAwakeFileDispatch: Remove dead code
// 06/08/10 19:50 NDP - ServerAwakeFileDispatch: fix bug: the check on package was done on full method name
// 27/09/10 20:10 NDP - use now JsonList to transport list of strings
// 28/08/10 18:30 NDP - actionListDirsOrFiles() sends content line per line on the output stream
// 24/09/10 12:50 NDP - ServerAwakeFileDispatch.actionInvokeRemoteMethod: String returned to client is now base64
// 25/09/10 09:10 NDP - ServerAwakeFileDispatch: call allowExecuteAfterAnalysis() before method execution
// 07/10/10 10:50 NDP : ServerAwakeFileDispatch: actionCall throws now SQLException
// 07/10/10 15:25 NDP : ServerAwakeFileDispatch: call free connection in finally
// 12/10/10 17:50 NDP : ServerAwakeFileDispatch: all actions on files are dispatched to AwakeFileActionManager
// 15/10/10 16:25 NDP : ServerAwakeFileDispatch: all JSON actions are done using JsonServerBuilder
// 15/10/10 16:25 NDP : ServerAwakeFileDispatch.call(): Add System.out.println("Call Exception: " + getStackTrace(e));
// 15/12/10 20:35 NDP : ServerAwakeFileDispatch.call(): Add System.out.println("Call Exception: " + getStackTrace(e));
// 16/01/11 20:35 NDP : ServerAwakeFileDispatch: filename = HtmlConverter.fromHtml(filename);
//
// 12/04/12 17:05 NDP : ServerAwakeFileDispatch: print class name, message & stack trace on servlet output stream
// 12/04/12 17:20 NDP : ServerAwakeFileDispatch: actionCall() throws now explicitely the reflection exceptions
// 28/04/12 19:25 NDP : ServerAwakeFileDispatch: log call() exceptions on System.out 
// 03/06/11 20:00 NDP : ServerAwakeFileDispatch.isTokenValid: rename param 
// 04/06/11 12:10 NDP : ServerAwakeFileDispatch: rename login to username
// 09/07/11 16:45 NDP : ServerAwakeFileDispatch: support all classes that have a TheClass(String s) contructor:
// 10/07/11 21:35 NDP : ServerAwakeFileDispatch: support call with AwakeCallable(NotAuthenticated) interface
// 16/07/11 13:15 NDP : ServerAwakeFileDispatch: remove old client call() security checks
// 07/10/11 13:20 NDP : ServerAwakeFileDispatch: use new JavaValueBuilder to decode params types & values in call
// 22/11/11 17:25 NDP : ServerAwakeFileDispatch: throw Exception on first getParameter(): means encryption fails
// 08/12/11 15:20 NDP : ServerAwakeFileDispatch: return result in base64 for call with parameter.CALL_ACTION
// 08/12/11 15:20 NDP : ServerAwakeFileDispatch: convert parameters from base64 for call with parameter.CALL_ACTION
// 10/12/11 21:15 NDP : ServerAwakeFileDispatch: remove all unnecessary base64 conversions
// 10/12/11 21:45 NDP : ServerAwakeFileDispatch: actionListDirsOrFiles: list dir/files were inverted. Fixed.
// 10/12/11 15:30 NDP : ServerAwakeFileDispatch: clean code if (action.equals(Action.CALL_ACTION))
// 03/01/12 12:20 NDP : ServerAwakeFileDispatch: in actionCall(): create a Connection only if required
//			(Allows to use Awake File without configuring JDBC pool)
// 04/01/12 15:15 NDP : ServerAwakeFileDispatch:  Run the runIfCallDisallowed() configured by the user
// 04/01/12 19:35 NDP : ServerAwakeFileDispatch: Add Tag.AWAKE_SECURITY to AwakeSecurityException

package org.awakefw.file.servlet;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.List;
import java.util.Vector;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.fileupload.servlet.ServletFileUpload;
import org.apache.commons.lang3.StringUtils;
import org.awakefw.commons.api.client.AwakeSecurityException;
import org.awakefw.commons.api.server.AwakeCommonsConfigurator;
import org.awakefw.file.api.server.AwakeFileConfigurator;
import org.awakefw.file.api.server.fileaction.AwakeFileActionManager;
import org.awakefw.file.api.util.AwakeDebug;
import org.awakefw.file.api.util.CallUtil;
import org.awakefw.file.api.util.HtmlConverter;
import org.awakefw.file.api.util.TokenUtil;
import org.awakefw.file.http.HttpTransfer;
import org.awakefw.file.servlet.json.JsonServerBuilder;
import org.awakefw.file.util.AwakeLogger;
import org.awakefw.file.util.JavaValueBuilder;
import org.awakefw.file.util.StringUtil;
import org.awakefw.file.util.Tag;
import org.awakefw.file.util.convert.HttpServletRequestConvertor;
import org.awakefw.file.util.parms.Action;
import org.awakefw.file.util.parms.Parameter;
import org.awakefw.file.util.parms.ReturnCode;

/**
 * @author Nicolas de Pomereu
 * 
 *         The method executeRequest() is to to be called from the
 *         ServerCallerRecv Servlet and Class. <br>
 *         It will execute a client side request with a ServerCaller.call()
 *         instance.
 * 
 */
public class ServerAwakeFileDispatch {
   
    private static boolean DEBUG = AwakeDebug
	    .isSet(ServerAwakeFileDispatch.class);

    // A space
    public static final String SPACE = " ";

    public static String CR_LF = System.getProperty("line.separator");

    /**
     * Constructor
     */
    public ServerAwakeFileDispatch() {

    }

    /**
     * 
     * Execute the dispatched request
     * 
     * @param request
     *            the http request
     * @param response
     *            the http response
     * @param awakeCommonsConfigurator
     *            the client commons configurator
     * @param awakeFileConfigurator
     *            the client configurator for files
     * @param awakeFileActionManager
     *            the client instance for File actions
     * @throws IOException
     *             if any Servlet Exception occurs
     */
    public void executeRequest(HttpServletRequest request,
	    HttpServletResponse response,
	    AwakeCommonsConfigurator awakeCommonsConfigurator,
	    AwakeFileConfigurator awakeFileConfigurator,
	    AwakeFileActionManager awakeFileActionManager) throws IOException {
	OutputStream out = null;

	try {
	    // System.out.println("ServerAwakeFileDispatch begin 1");

	    // Immediate catch if we are asking a file upload, because
	    // parameters are
	    // in unknown sequence. We know it's a upload action if it's mime
	    // multipart
	    if (ServletFileUpload.isMultipartContent(request)) {
		ServerFileUploadAction dispatchAction = new ServerFileUploadAction();
		dispatchAction.executeAction(request, response,
			awakeCommonsConfigurator, awakeFileConfigurator,
			awakeFileActionManager);
		return;
	    }

	    debug("ServerAwakeFileDispatch begin 2");

	    // The action & filename (for file size ask)
	    String action = null;
	    
	    // We must trap the IllegalArgumentException to rethrow properly to client
	    // This happens if there is an encryption problem
	    try {
		action = request.getParameter(Parameter.ACTION);
	    } catch (IllegalArgumentException e) {
		out = response.getOutputStream();
		throw e;
	    }

	    action = this.getTrimValue(action);

	    // Special action for Login, because Token does not exists and must
	    // be built
	    if (action.equals(Action.LOGIN_ACTION)
		    || action.equals(Action.BEFORE_LOGIN_ACTION)) {
		ServerLoginAction dispatchAction = new ServerLoginAction();
		dispatchAction.executeAction(request, response,
			awakeCommonsConfigurator, action);
		return;
	    }

	    out = response.getOutputStream();

	    // Only if there is a call action, we may execute authorized classes
	    // without authentication/login
	    if (action.equals(Action.CALL_ACTION) || action.equals(Action.CALL_ACTION_HTML_ENCODED )) {
		// The class name
		String methodName = request.getParameter(Parameter.METHOD_NAME);
		methodName = this.getTrimValue(methodName);

		String className = StringUtils.substringBeforeLast(methodName,
			".");
		Class<?> c = Class.forName(className);
		CallUtil callUtil = new CallUtil(c);
		boolean callAllowed = callUtil.isCallableNotAuthenticated();

		if (callAllowed) {
		    if (action.equals(Action.CALL_ACTION) || action.equals(Action.CALL_ACTION_HTML_ENCODED)) {
			actionCall(request, awakeCommonsConfigurator,
				awakeFileConfigurator, out, null);
		    } 
			
		    return;
		}
	    }

	    // For all other actions, we check the parameters

	    // The username (used for token re-compilation)
	    String username = request.getParameter(Parameter.LOGIN);
	    username = this.getTrimValue(username);

//	    // For old call():
//	    try {
//		username = StringUtil.fromBase64(username);
//	    } catch (Exception e) {
//	    } // The login may be in clear

	    // Authentication Token with SHA-1(login + secret value)
	    String token = request.getParameter(Parameter.TOKEN);
	    token = this.getTrimValue(token);

	    if (!isTokenValid(username, token, awakeCommonsConfigurator)) {
		// writeLine(out, HttpTransfer.SEND_OK + SPACE +
		// ReturnCode.INVALID_LOGIN_OR_PASSWORD);
		
		debug("username: " + username);
		debug("token   : " + token);
		
		writeLine(out, HttpTransfer.SEND_OK);
		writeLine(out, ReturnCode.INVALID_LOGIN_OR_PASSWORD);

		return;
	    }

	    // The filename
	    String filename = request.getParameter(Parameter.FILENAME);
	    filename = this.getTrimValue(filename);

	    // Treat all actions
	    if (action.equals(Action.DELETE_FILE_ACTION)) {
		boolean result = awakeFileActionManager.deleteFile(
			awakeFileConfigurator, username, filename);

		writeLine(out, HttpTransfer.SEND_OK);
		writeLine(out, "" + result);
	    } else if (action.equals(Action.MKDIRS_ACTION)) {
		boolean result = awakeFileActionManager.mkdirs(
			awakeFileConfigurator, username, filename);

		writeLine(out, HttpTransfer.SEND_OK);
		writeLine(out, "" + result);
	    } else if (action.equals(Action.LIST_DIRS_IN_DIR_ACTION)
		    || action.equals(Action.LIST_FILES_IN_DIR_ACTION)) {
		actionListDirsOrFiles(awakeFileConfigurator,
			awakeFileActionManager, username, filename, action, out);

	    } else if (action.equals(Action.GET_FILE_LENGTH_ACTION)) {
		long result = actionGetListFileLength(awakeFileConfigurator,
			awakeFileActionManager, username, filename);

		writeLine(out, HttpTransfer.SEND_OK);
		writeLine(out, Long.toString(result));
	    } else if (action.equals(Action.DOWNLOAD_FILE_ACTION)) {

		boolean result = awakeFileActionManager.downloadFile(out,
			awakeFileConfigurator, username, filename);

		if (!result) {
		    // Impossible to find the file on server
		    // writeLine(out, HttpTransfer.SEND_OK);
		    // writeLine(out, Tag.FileNotFoundException);
		    throw new FileNotFoundException(
			    "File not found on remote server: " + filename);
		}
	    } else if (action.equals(Action.CALL_ACTION) || action.equals(Action.CALL_ACTION_HTML_ENCODED)) {
		try {
		    actionCall(request, awakeCommonsConfigurator,
			    awakeFileConfigurator, out, username);
		} catch (Exception e) {
		    e.printStackTrace(System.out);
		    throw e;
		}
							
	    } else {
		throw new IllegalArgumentException("Invalid Client Action: "
			+ action);
	    }

	    return;
	} catch (Exception e) {
	    // Always for Tomcat logging!
	    // e.printStackTrace();

	    if (DEBUG) {
		e.printStackTrace(System.out);
	    }

	    writeLine(out, HttpTransfer.SEND_FAILED);
	    writeLine(out, e.getClass().getName()); // Exception class name
	    writeLine(out, e.getMessage()); // Exception message
	    writeLine(out, getStackTrace(e)); // stack trace
	}
    }

    /**
     * Gets the specified exception stack trace as a String.
     * 
     * @param eThrown
     *            the exception
     * @return the exception stack trace
     */

    public static final String getStackTrace(Exception eThrown) {
	if (eThrown == null)
	    return null;

	try {
	    ByteArrayOutputStream baos = new ByteArrayOutputStream();
	    PrintWriter pw = new PrintWriter(baos);
	    eThrown.printStackTrace(pw);
	    pw.close();
	    String sStack = new String(baos.toByteArray());
	    return sStack;
	} catch (Exception eProb) {
	    // hope you can see that in your java console
	    System.out.println("While reporting exception : "
		    + eProb.toString());
	    return eThrown.toString();
	}
    }

    /**
     * Write a line of string on the servlet output stream. Will add the
     * necessary CR_LF
     * 
     * @param out
     *            the servlet output stream
     * @param s
     *            the string to write
     * @throws IOException
     */
    private void writeLine(OutputStream out, String s) throws IOException {
	out.write((s + CR_LF).getBytes());
    }

    /**
     * Check the validity of the (username, token) pair <br>
     * Will return false if the operation is now allowed!
     * 
     * @param username
     *            the username to check
     * @param token
     *            the associated token with the username
     * @param awakeCommonsConfigurator
     *            the user configuration
     * 
     * @return true if the pair (username, token) is verified and ok.
     * @throws Exception
     */
    private boolean isTokenValid(String username, String token,
	    AwakeCommonsConfigurator awakeCommonsConfigurator) throws Exception {

	// OK! Now build a token with SHA-1(login + secret)
	String tokenRecomputed = TokenUtil.computeToken(username,
		awakeCommonsConfigurator.addSecretForAuthToken());

	if (token == null || !token.equals(tokenRecomputed)) {
	    debug("login          : " + username + ":");
	    debug("token          : " + token + ":");
	    debug("tokenRecomputed: " + tokenRecomputed + ":");
	    return false;
	}

	return true;
    }

    /**
     * Action: List all sub-directories in a directory
     * 
     * @param awakeFileConfigurator
     *            the user configuration
     * @param awakeFileActionManager
     *            the user file manager
     * @param username
     *            the client login
     * @param filename
     *            the full path of directory to list files od directory
     * @param out
     *            the output stream
     * @throws IOException
     */
    private void actionListDirsOrFiles(
	    AwakeFileConfigurator awakeFileConfigurator,
	    AwakeFileActionManager awakeFileActionManager, String username,
	    String filename, String action, OutputStream out)
	    throws IOException {
	debug("actionListDirsOrFiles: " + action);

	List<File> files = null;

	if (action.equals(Action.LIST_FILES_IN_DIR_ACTION)) {
	    files = awakeFileActionManager.listFiles(awakeFileConfigurator,
		    username, filename);
	} else if (action.equals(Action.LIST_DIRS_IN_DIR_ACTION)) {
	    files = awakeFileActionManager.listDirectories(
		    awakeFileConfigurator, username, filename);
	}
	else {
	    throw new IllegalArgumentException(Tag.AWAKE_FAIL + "Invalid list dirs/files action: " + action);
	}
	    

	if (files == null || files.size() == 0) {
	    writeLine(out, HttpTransfer.SEND_OK);
	    writeLine(out, null);
	    return;
	}

	writeLine(out, HttpTransfer.SEND_OK);

	for (File file : files) {
	    	    
	    String theFile = file.getName();
	    theFile = HtmlConverter.toHtml(theFile);

	    writeLine(out, theFile);
	}
    }

    /**
     * Action: delete a file list
     * 
     * @param awakeFileConfigurator
     * @param awakeFileActionManager
     * 
     * @param filename
     *            the filelist
     */
    private long actionGetListFileLength(
	    AwakeFileConfigurator awakeFileConfigurator,
	    AwakeFileActionManager awakeFileActionManager, String username,
	    String filename) throws IOException {
	debug("Action.GET_FILE_LENGTH_ACTION");
	long result = 0;

	// We have in fact a list of files
	// List<String> files = UtilBuilder.toListOfStrings(filename);
	List<String> files = JsonServerBuilder.fromJson(filename);

	for (String theFilename : files) {
	    //theFilename = HtmlConverter.fromHtml(theFilename);

	    result += awakeFileActionManager.getFileLength(
		    awakeFileConfigurator, username, theFilename);
	}

	return result;
    }

    /**
     * return a "" empty string if null or a trimed stgring if not null
     * 
     * @param action
     *            the input string
     * @return the empty stgring or trimed string
     */
    private String getTrimValue(String string) {
	if (string == null) {
	    return "";
	} else {
	    return string.trim();
	}
    }

    /**
     * 
     * Calls a remote method from the PC <br>
     * Please note that all invocation are trapped and routed as code string to
     * the PC.
     * 
     * @param request
     *            the http request
     * @param awakeCommonsConfigurator
     *            the commons configurator defined by the user
     * @param awakeFileConfigurator
     *            the file configurator defined by the user
     * @param out
     *            the servlet output stream
     * @param username
     *            the client login (for security check)
     * 
     * 
     * @throws IOException
     *             all Awake, network, etc. errors
     * @throws ClassNotFoundException
     * @throws IllegalAccessException
     * @throws InstantiationException
     * @throws NoSuchMethodException
     * @throws AwakeSecurityException
     * @throws InvocationTargetException
     * @throws IllegalArgumentException
     */
    private void actionCall(HttpServletRequest request,
	    AwakeCommonsConfigurator awakeCommonsConfigurator,
	    AwakeFileConfigurator awakeFileConfigurator, OutputStream out,
	    String username) throws SQLException, IOException,
	    AwakeSecurityException, ClassNotFoundException,
	    InstantiationException, IllegalAccessException,
	    NoSuchMethodException, IllegalArgumentException,
	    InvocationTargetException {

	Connection connection = null;

	try {
	    
	    debug("in actionCall");
	    
	    // The method name
	    String methodName = request.getParameter(Parameter.METHOD_NAME);

	    // The parms name
	    String paramsTypes = request.getParameter(Parameter.PARAMS_TYPES);
	    String paramsValues = request.getParameter(Parameter.PARAMS_VALUES);

	    // Make sure all values are not null and trimed

	    methodName = this.getTrimValue(methodName);
	    paramsTypes = this.getTrimValue(paramsTypes);
	    paramsValues = this.getTrimValue(paramsValues);

	    if (request instanceof HttpServletRequestConvertor) {
		debug("request instanceof HttpServletRequestConvertor");
	    } 
	    else {
		debug("request NOT instanceof HttpServletRequestConvertor");
	    }
			    
	    debug("actionInvokeRemoteMethod:methodName       : " + methodName);

	    String className = StringUtils.substringBeforeLast(methodName, ".");
	    Class<?> c = Class.forName(className);
	    CallUtil callUtil = new CallUtil(c);
	    boolean callAllowed = callUtil.isCallable();

	    if (!callAllowed) {
		throw new AwakeSecurityException(Tag.AWAKE_SECURITY
			+ "Class is forbiden for remote call: " + className);
	    }
	    
	    String action = request.getParameter(Parameter.ACTION);
	    
	    // Legacy Action.CALL_ACTION call with Base64 conversion
	    // Corresponds to AwakeFileSession.setUseBase64EncodingForCall() setting
	    // on client side
	    if (action.equals(Action.CALL_ACTION)) {
		paramsTypes = StringUtil.fromBase64(paramsTypes);
		paramsValues = StringUtil.fromBase64(paramsValues);
	    }	    
	    
	    debug("paramsTypes     : " + paramsTypes);
	    debug("paramsValues    : " + paramsValues);
	    
	    List<String> listParamsTypes = JsonServerBuilder
		    .fromJson(paramsTypes);
	    List<String> listParamsValues = JsonServerBuilder
		    .fromJson(paramsValues);

	    debug("actionInvokeRemoteMethod:listParamsTypes      : "
		    + listParamsTypes);
	    debug("actionInvokeRemoteMethod:listParamsValues     : "
		    + listParamsValues);

	    Class<?>[] argTypes = new Class[listParamsTypes.size()];
	    Object[] values = new Object[listParamsValues.size()];

	    List<Object> valuesList = new Vector<Object>();
	    for (int i = 0; i < listParamsTypes.size(); i++) {

		String value = listParamsValues.get(i);
		String javaType = listParamsTypes.get(i);

		JavaValueBuilder javaValueBuilder = new JavaValueBuilder(
			javaType, value);
		argTypes[i] = javaValueBuilder.getClassOfValue();
		values[i] = javaValueBuilder.getValue();
		
		// Special treatement if argTypes[i] is a Connection
		if (argTypes[i] == Connection.class) {
		    try {
			connection = awakeCommonsConfigurator.getConnection();
			values[i] = connection;
		    } catch (SQLException e) {
			throw new IOException(e);
		    }
		}
		
		valuesList.add(values[i]);
	    }

	    if (!awakeFileConfigurator.allowCallAfterAnalysis(username,
		    connection, methodName, valuesList)) {
		
		String ipAddress = request.getRemoteAddr();
		
		// Run the runIfCallDisallowed() configured by the user
		awakeFileConfigurator.runIfCallDisallowed(username, connection, ipAddress, methodName, valuesList);
		
		throw new AwakeSecurityException(Tag.AWAKE_SECURITY + 
			"Method not authorized for execution by Security Checker : "
				+ methodName + " parameters: "
				+ valuesList.toString());
	    }

	    Object theObject = c.newInstance();
	    String rawMethodName = StringUtils.substringAfterLast(methodName,
		    ".");

	    // Invoke the method
	    Method main = c.getDeclaredMethod(rawMethodName, argTypes);
	    Object resultObj = main.invoke(theObject, values);
	    String result = null;

	    if (resultObj != null)
		result = resultObj.toString();

	    debug("result before conversion: " + result);
	    
	    if (result != null) {

		// Legacy Action.CALL_ACTION call with Base64 conversion
		// Corresponds to AwakeFileSession.setUseBase64EncodingForCall()
		// setting on client side	
		if (action.equals(Action.CALL_ACTION)) {
		    result = StringUtil.toBase64(result);
		} else if (action.equals(Action.CALL_ACTION_HTML_ENCODED)) {
		    result = HtmlConverter.toHtml(result);
		}
		else {
		    throw new IllegalArgumentException("call action is invalid: " + action);
		}
	    }
	    
	    debug("actionInvokeRemoteMethod:result: " + result);
	    
	    writeLine(out, HttpTransfer.SEND_OK);
	    writeLine(out, result);
	} finally {
	    if (connection != null) {
		connection.close();
	    }
	}
    }    
    
    private void debug(String s) {
	if (DEBUG) {
	    AwakeLogger.log(s);
	}
    }
}
