/*
 * Awake File: Easy file upload & download through HTTP with Java
 * Awake SQL: Remote JDBC access through HTTP.                                    
 * Copyright (C) 2011, Kawan Softwares S.A.S.
 * (http://www.awakeframework.org). All rights reserved.                                
 *                                                                               
 * Awake File/SQL is free software; you can redistribute it and/or                 
 * modify it under the terms of the GNU Lesser General Public                    
 * License as published by the Free Software Foundation; either                  
 * version 2.1 of the License, or (at your option) any later version.            
 *                                                                               
 * Awake File/SQL is distributed in the hope that it will be useful,               
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             
 * Lesser General Public License for more details.                               
 *                                                                               
 * You should have received a copy of the GNU Lesser General Public              
 * License along with this library; if not, write to the Free Software           
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301  USA
 *
 * Any modifications to this file must keep this entire header
 * intact.
 */

//BasicNameValuePairEncryptor.java
//Copyright (c) Kawan Softwares S.A.S, 2011
//
//Last Updates: 
// 16 nov. 2010 19:13:15 Nicolas de Pomereu
// 20/11/10 21:00 NDP : BasicNameValuePairEncryptor: do not encrypt if password is null!
// 09/03/11 20:05 NDP - BasicNameValuePairEncryptor: do not encrypt for Parameter.STATEMENT_HOLDER
// 08/12/11 19:35 NDP : BasicNameValuePairConvertor: new name & html conversion (ex: BasicNameValuePairEncryptor)
// 10/12/11 12:45 NDP : BasicNameValuePairConvertor: pass parameter VERSION with Awake File version
// 12/12/11 13:45 NDP : BasicNameValuePairConvertor: no more reference to Action.CALL_ACTION
// 12/12/11 21:30 NDP : BasicNameValuePairConvertor: encryption is done before to Html conversion


package org.awakefw.file.util.convert;

import java.io.IOException;
import java.util.List;
import java.util.Vector;

import org.apache.http.message.BasicNameValuePair;
import org.awakefw.commons.api.client.HttpProtocolParameters;
import org.awakefw.file.api.util.AwakeDebug;
import org.awakefw.file.api.util.HtmlConverter;
import org.awakefw.file.util.parms.Parameter;
import org.awakefw.file.version.AwakeFileVersionValues;

public class BasicNameValuePairConvertor {
    /** Debug flag */
    private static boolean DEBUG = AwakeDebug
	    .isSet(BasicNameValuePairConvertor.class);

    /** The request params */
    private List<BasicNameValuePair> requestParams = null;

    /** The http protocol parameters */
    private HttpProtocolParameters httpProtocolParameters = null;

    /**
     * Constructor
     * 
     * @param requestParams
     *            the request parameters list of BasicNameValuePair
     */
    public BasicNameValuePairConvertor(List<BasicNameValuePair> requestParams, 
	    	HttpProtocolParameters httpProtocolParameters) {
	this.requestParams = requestParams;
	this.httpProtocolParameters = httpProtocolParameters;
    }

    /**
     * @return the convert & may be encrypted parameters
     */
    public List<BasicNameValuePair> convert() throws IOException {
	List<BasicNameValuePair> requestParamsConverted = new Vector<BasicNameValuePair>();

	for (BasicNameValuePair basicNameValuePair : requestParams) {
	    String paramName = basicNameValuePair.getName();	    
	    String paramValue = basicNameValuePair.getValue();

	    // Don't do it for STATEMENT_HOLDER ==> already has it's own Html
	    // conversion and encryption
	    if (! paramName.equals(Parameter.STATEMENT_HOLDER))
	    {
		paramValue = HtmlConverter.toHtml(paramValue);
		paramValue = encryptValue(paramName, paramValue);
	    }
		    	  
	    debug("converted param name : " + paramName);
	    debug("converted param value: " + paramValue);

	    BasicNameValuePair basicNameValuePairEnc = new BasicNameValuePair(
		    paramName, paramValue);
	    requestParamsConverted.add(basicNameValuePairEnc);
	}
	
	// Add the Version
	BasicNameValuePair basicNameValuePairEnc = new BasicNameValuePair(
		Parameter.VERSION, AwakeFileVersionValues.VERSION);
	requestParamsConverted.add(basicNameValuePairEnc);

	return requestParamsConverted;
    }

    /**
     * Encrypt the parameter value
     * 
     * @param paramName
     *            the parameter name
     * @param paramValue
     *            the parameter value
     * 
     * @return the encrypted parameter value
     * 
     * @throws IOException
     *             if any exception occurs
     */
    private String encryptValue(String paramName, String paramValue)
	    throws IOException {
	
	if (httpProtocolParameters == null) {
	    return paramValue;
	}
	
	char [] password = httpProtocolParameters.getEncryptionPassword();

	if (password == null || password.length <= 1) {
	    return paramValue;
	}
	    	
	try {
	    paramValue = Pbe.AWAKE_ENCRYPTED
		    + new Pbe().encryptToHexa(paramValue, password);
	} catch (Exception e) {
	    String message = "Impossible to encrypt the value of the parameter "
		    + paramName;

	    throw new IOException(message, e);
	}
	return paramValue;
    }

    /**
     * Displays the given message if DEBUG is set.
     * 
     * @param s
     *            the debug message
     */

    private static void debug(String s) {
	if (DEBUG) {
	    System.out.println(s);
	}
    }

}
